import math
import torch
from torch import nn

_sqrt_2 = math.sqrt(2)


class Quadratic(nn.Module):
    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        super().__init__()
        self.linear = nn.Linear(in_features * 2, out_features, bias=bias)

    def forward(self, x: torch.Tensor):
        x_pow2 = (x.pow(2) - 1.0) / _sqrt_2
        x_expanded = torch.cat([x, x_pow2], dim=1)
        return self.linear(x_expanded)



import logging
import math
import torch
from torch import autograd
from neural_commons.cf_nn import LRStrategy

_logger = logging.getLogger("neural-commons")
_max_base_factor = 1.5
_min_base_factor = 0.5
_sd_mean_ts = 0.01


class AdaptiveLRS(LRStrategy):
    def __init__(self, min_loss: float = 0, deficit_factor: float = 1.0):
        super().__init__()
        self.min_loss = min_loss
        self.deficit_factor = deficit_factor
        self.warned_low_loss = False

    def requires_second_derivative(self) -> bool:
        return True

    def get_residual(self, is_final: bool, output: torch.Tensor, loss: torch.Tensor, grad1: torch.Tensor,
                     final_output_size: int) -> torch.Tensor:
        grad2 = autograd.grad(grad1.sum(), output)[0]
        if self.is_zero(grad2) or not self.is_flat(grad2):
            return self.get_slide_residual(loss, grad1)
        else:
            return self.get_newton_method_residual(grad1, grad2)

    def get_slide_residual(self, loss: torch.Tensor, grad1: torch.Tensor, eps=1e-15):
        deficit = loss.item() - self.min_loss
        if deficit < 0:
            deficit = 0
            if not self.warned_low_loss:
                _logger.warning(f"SlideLRS: Loss ({loss.item()}) is less than min_loss ({self.min_loss}).")
                self.warned_low_loss = True
        deficit *= self.deficit_factor
        denominator = torch.clamp(torch.sum(grad1.pow(2)), min=eps)
        lr = (deficit / denominator).item()
        return -lr * grad1

    @staticmethod
    def get_newton_method_residual(grad1: torch.Tensor,
                                   grad2: torch.Tensor):
        return -grad1 / grad2

    @staticmethod
    def is_zero(tensor: torch.Tensor) -> bool:
        return torch.all(torch.abs(tensor) <= 1e-20).item()

    @staticmethod
    def is_flat(tensor: torch.Tensor) -> bool:
        if torch.numel(tensor) < 3 or torch.any(tensor <= 0).item():
            return False
        t_mean = torch.mean(tensor).item()
        t_std = torch.std(tensor).item()
        return t_std / t_mean <= _sd_mean_ts

    @staticmethod
    def get_base_lr(is_final: bool, output: torch.Tensor,
                    final_output_size: int):
        batch_size = output.size(0)
        output_size = math.prod(output.shape[1:])
        fos = output_size if is_final else final_output_size
        base_result = 0.5 * batch_size * fos
        if is_final:
            lr = base_result
        else:
            factor_min = 1.0
            factor_max = 12.0
            partial_os = output_size / 2
            max_w = partial_os / (partial_os + final_output_size)
            factor = factor_min * (1 - max_w) + factor_max * max_w
            lr = factor * base_result
        return lr



import torch
from torch import nn


class ParametricReLU(nn.Module):
    def __init__(self, input_shape: tuple[int, ...]):
        super(ParametricReLU, self).__init__()
        self.input_shape = input_shape
        self.slopes = nn.Parameter(torch.ones(input_shape))
        setattr(self.slopes, "__skip_transplant__", True)

    def forward(self, x: torch.Tensor):
        positive = x >= 0
        return positive * x + (~positive) * x * self.slopes[None, :]



import logging
import torch
from neural_commons.cf_nn import LRStrategy
from neural_commons.helpers.torch_helper import view_dims, dims_from, num_dims

_logger = logging.getLogger("neural-commons")


class SlideLRS(LRStrategy):
    def __init__(self, min_loss: float, deficit_factor: float = 1.0):
        super().__init__()
        if deficit_factor <= 0 or deficit_factor > 1:
            raise ValueError("final_factor must be in ]0, 1].")
        self.deficit_factor = deficit_factor
        self.min_loss = min_loss
        self.warned_low_loss = False

    def requires_second_derivative(self) -> bool:
        return False

    def get_residual(self, is_final: bool, output: torch.Tensor, loss: torch.Tensor, grad1: torch.Tensor,
                     final_output_size: int, eps=1e-30):
        # loss shape: (batch_size)
        deficit = loss - self.min_loss
        if torch.any(deficit < 0).item():
            deficit = torch.clamp(deficit, min=0)
            if not self.warned_low_loss:
                _logger.warning(f"SlideLRS: Loss contains a value less than min_loss ({self.min_loss}).")
                self.warned_low_loss = True
        deficit *= self.deficit_factor
        denominator = torch.clamp(torch.sum(grad1.pow(2), dim=dims_from(grad1)), min=eps)
        lr = deficit / denominator
        # lr shape: (batch_size)
        return -view_dims(lr, num_dims(grad1)) * grad1



import copy
import torch
from torch import nn
from neural_commons.modules import SMoEMixIn


class SMoE(nn.Module, SMoEMixIn):
    cluster_centroids: torch.Tensor

    def __init__(self, module: nn.Module, in_features: int, out_features: int,
                 num_experts: int = 64, ensemble_size: int = 3):
        if ensemble_size <= 0:
            raise ValueError(f"Invalid ensemble_size={ensemble_size}")
        if num_experts <= 0:
            raise ValueError(f"Invalid num_experts={num_experts}")
        if num_experts % 2 != 0:
            raise ValueError(f"num_experts(={num_experts}) must be an even number")
        if ensemble_size > num_experts:
            raise ValueError(f"ensemble_size={ensemble_size} > num_experts={num_experts}")
        super(SMoE, self).__init__()
        self.out_features = out_features
        self.experts = nn.ModuleList([copy.deepcopy(module) for _ in range(num_experts)])
        self.num_experts = num_experts
        self.ensemble_size = ensemble_size
        self.d_scale = nn.Parameter(torch.FloatTensor([[[0.1]]]))
        vectors = self.get_orthogonal_vectors(num_experts // 2, in_features)
        centroids = (torch.cat([vectors, -vectors], dim=0)
                     .detach().unsqueeze(0))
        # centroids: (1, num_experts, in_features,)
        self.register_buffer("cluster_centroids", centroids)

    def forward(self, x: torch.Tensor):
        es = self.ensemble_size
        # x: (batch_size, in_features)
        scaled_diff = self.d_scale * (x[:, None, :] - self.cluster_centroids)
        distance_sq = torch.mean(scaled_diff.pow(2), dim=2)
        # distance_sq: (batch_size, num_experts)
        sort_indexes = torch.argsort(distance_sq, dim=1)
        # sort_indexes: (batch_size, num_experts)
        picked_indexes = sort_indexes[:, :es]
        # picked_indexes: (batch_size, ensemble_size)
        select_distances = distance_sq[:, :es]
        select_weights = 1.0 / (1.0 + select_distances)
        # select_weights: (batch_size, ensemble_size)
        weight_sums = torch.sum(select_weights, dim=1, keepdim=True)
        select_weights /= weight_sums
        unique_weight_indexes = torch.unique(picked_indexes)
        batch_size = x.size(0)
        expert_out = torch.empty((batch_size, es, self.out_features), device=x.device)
        exec_indexes = []
        exec_results = []
        for wi in unique_weight_indexes:
            indexes = torch.where(picked_indexes == wi)
            row_idx, _ = indexes
            x_subset = x[row_idx]
            module = self.experts[wi.item()]
            result = module(x_subset)
            exec_indexes.append(indexes)
            exec_results.append(result)
        for (row_idx, e_idx), output in zip(exec_indexes, exec_results):
            expert_out[row_idx, e_idx, :] = output
        return torch.sum(expert_out * select_weights[:, :, None], dim=1)



import unittest
import torch
from torch import nn

from neural_commons.cf_nn import CFLinear, CFOptimizer, CFConv2d
from neural_commons.cf_nn.tests.utils import IrregularLoss, InvGaussianLoss
from neural_commons.modules import RndProjection, ArgMax, MAELoss, View
from neural_commons.modules.Plus import Plus


class TestCFOptimizer(unittest.TestCase):
    def test_regression_with_bias(self):
        in_features = 32
        out_features = 24
        batch_size = 200
        ref_m = nn.Sequential(
            nn.Linear(in_features, out_features, bias=False),
            Plus(5),
        )
        opt_m = nn.Sequential(
            CFLinear(in_features, out_features, bias=True),
        )
        loss_fn = nn.MSELoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.3)

    def test_regression_without_bias(self):
        in_features = 25
        out_features = 44
        batch_size = 200
        ref_m = nn.Linear(in_features, out_features, bias=False)
        opt_m = nn.Sequential(
            CFLinear(in_features, out_features, bias=False),
        )
        loss_fn = nn.MSELoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features,
                                                 num_epochs=1)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.3)

    def test_conv2d_with_bias(self):
        in_channels = 16
        out_channels = 24
        batch_size = 200
        width = height = 8
        in_features = width * height * in_channels
        ref_m = nn.Sequential(
            View(-1, in_channels, height, width),
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, bias=False),
            View(-1, out_channels * width * height),
            Plus(-3.0),
        )
        opt_m = nn.Sequential(
            View(-1, in_channels, height, width),
            CFConv2d(in_channels, out_channels, kernel_size=3, padding=1, bias=True),
            View(-1, out_channels * width * height),
        )
        loss_fn = nn.MSELoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.3)

    def test_conv2d_without_bias(self):
        in_channels = 16
        out_channels = 24
        batch_size = 200
        width = height = 8
        in_features = width * height * in_channels
        ref_m = nn.Sequential(
            View(-1, in_channels, height, width),
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, bias=False),
            View(-1, out_channels * width * height),
        )
        opt_m = nn.Sequential(
            View(-1, in_channels, height, width),
            CFConv2d(in_channels, out_channels, kernel_size=3, padding=1, bias=False),
            View(-1, out_channels * width * height),
        )
        loss_fn = nn.MSELoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.3)

    def test_multi_layer(self):
        in_features = 24
        out_features = 32
        batch_size = 200
        ref_m = nn.Linear(in_features, out_features, bias=False)
        opt_m = nn.Sequential(
            CFLinear(in_features, 128),
            RndProjection(128, out_features),
        )
        loss_fn = nn.MSELoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.3)

    def test_mae_loss(self):
        in_features = 32
        out_features = 24
        batch_size = 500
        ref_m = nn.Linear(in_features, out_features, bias=True)
        opt_m = nn.Sequential(
            CFLinear(in_features, out_features, bias=True),
        )
        loss_fn = MAELoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features,
                                                 num_epochs=3)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.5)

    def test_ce_loss(self):
        in_features = 34
        out_features = 25
        batch_size = 100
        ref_m = nn.Sequential(
            nn.Linear(in_features, out_features, bias=True),
            ArgMax(),
        )
        opt_m = nn.Sequential(
            CFLinear(in_features, out_features, bias=True),
        )
        loss_fn = nn.CrossEntropyLoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.3)

    def test_inv_gaussian_loss(self):
        in_features = 32
        out_features = 24
        batch_size = 200
        ref_m = nn.Linear(in_features, out_features, bias=True)
        opt_m = nn.Sequential(
            CFLinear(in_features, out_features, bias=True),
        )
        loss_fn = InvGaussianLoss()
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features,
                                                 num_epochs=3)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.5)

    def test_irregular_loss(self):
        in_features = 34
        out_features = 30
        batch_size = 200
        ref_m = nn.Linear(in_features, out_features, bias=True)
        opt_m = nn.Sequential(
            CFLinear(in_features, out_features, bias=True),
        )
        loss_fn = IrregularLoss(out_features)
        loss_value_1, loss_value_2 = self._train(ref_m, opt_m, loss_fn, batch_size, in_features,
                                                 num_epochs=3)
        print(f"Initial loss value: {loss_value_1}")
        print(f"Final loss value: {loss_value_2}")
        self.assertLess(loss_value_2, loss_value_1 * 0.8)

    def _train(self, ref_m: nn.Module,
               opt_m: nn.Module, loss_fn: nn.Module, batch_size: int, in_features: int,
               num_epochs: int = 1):
        x = torch.randn((batch_size, in_features))
        label = ref_m(x)
        optimizer = CFOptimizer(opt_m.cf_modules())
        y1 = opt_m(x)
        loss_value_1 = loss_fn(y1, label).item()
        for e in range(num_epochs):
            def _fn() -> torch.Tensor:
                y = opt_m(x)
                _loss = loss_fn(y, label)
                return _loss

            optimizer.enqueue_data(_fn)
            if optimizer.queued_batch_size >= 1:
                optimizer.step()
        pred_y_2 = opt_m(x)
        loss = loss_fn(pred_y_2, label)
        loss_value_2 = torch.mean(loss).item()
        return loss_value_1, loss_value_2,



import datetime
import math
import os

import torch
from neural_commons.cf_nn import LRStrategy


class MSELossLRS(LRStrategy):
    def __init__(self):
        super().__init__()

    def requires_second_derivative(self) -> bool:
        return False

    def get_residual(self, is_final: bool, output: torch.Tensor, loss: torch.Tensor, grad1: torch.Tensor,
                     final_output_size: int):
        batch_size = output.size(0)
        output_size = math.prod(output.shape[1:])
        fos = output_size if is_final else final_output_size
        base_result = 0.5 * batch_size * fos
        if is_final:
            lr = base_result
        else:
            factor_min = 1.0
            factor_max = 10.0
            partial_os = output_size / 2
            max_w = partial_os / (partial_os + final_output_size)
            factor = factor_min * (1 - max_w) + factor_max * max_w
            lr = factor * base_result
        return -lr * grad1



import math
import torch
import torch.nn.functional as F
from torch import nn

from neural_commons.q_nn import ParamModule

_sqrt_2 = math.sqrt(2)


class QLinear(nn.Module):
    __constants__ = ["in_features", "out_features"]

    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        super().__init__()
        inv_scale = math.sqrt(in_features)
        if bias:
            inv_scale *= _sqrt_2
        weight = torch.randn((out_features, in_features)) / inv_scale
        bias_param = torch.randn((out_features,)) / _sqrt_2 if bias else None
        p_tensors = [weight]
        if bias_param is not None:
            p_tensors.append(bias_param)
        self.p_module = ParamModule(p_tensors)
        self.bias = bias
        self.in_features = in_features
        self.out_features = out_features

    def forward(self, x: torch.Tensor):
        if self.bias:
            w, b = self.p_module()
        else:
            w = self.p_module()[0]
            b = None
        return F.linear(x, w, b)

    def extra_repr(self) -> str:
        return f'in_features={self.in_features}, out_features={self.out_features}, bias={self.bias}'



import torch
from typing import Iterable
from .ParamModule import ParamModule
from .QGRFOptimizer import QGRFOptimizer
from .QLinear import QLinear


def _p_modules(self: torch.nn.Module) -> Iterable['ParamModule']:
    return (m for m in self.modules() if isinstance(m, ParamModule))


torch.nn.Module.param_modules = _p_modules



import torch
from torch import nn

from neural_commons.modules import Permute


class ConvMixer2d(nn.Module):
    def __init__(self, num_channels: int, width: int, height: int,
                 hidden_factor=3, kernel_size=3, hidden_s=10):
        super().__init__()
        self.height = height
        self.width = width
        self.num_channels = num_channels
        padding = kernel_size // 2
        conv_params = dict(kernel_size=kernel_size, padding=padding)
        hidden_w = width * hidden_factor
        hidden_h = height * hidden_factor
        self.mixer = nn.Sequential(
            Permute(0, 2, 1, 3),
            # shape: (batch_size, cell_img_size, hidden1, cell_img_size)
            nn.Conv2d(width, hidden_w, **conv_params),
            nn.ELU(),
            nn.Conv2d(hidden_w, width, **conv_params),
            Permute(0, 3, 1, 2),
            # shape: (batch_size, cell_img_size, cell_img_size, hidden1)
            nn.Conv2d(height, hidden_h, **conv_params),
            nn.ELU(),
            nn.Conv2d(hidden_h, height, **conv_params),
            Permute(0, 3, 2, 1),
        )
        self.select = nn.Sequential(
            nn.Conv2d(num_channels * 2, hidden_s, **conv_params),
            nn.ELU(),
            nn.Conv2d(hidden_s, 1, **conv_params),
            nn.Sigmoid(),
        )

    def forward(self, x: torch.Tensor):
        # x: (batch_size, channels, width, height)
        mixed_x = self.mixer(x)
        combined_x = torch.cat([x, mixed_x], dim=1)
        select_weight = self.select(combined_x)
        return mixed_x * select_weight + x * (1 - select_weight)


if __name__ == '__main__':
    _x = torch.randn((3, 16, 14, 18))
    m = ConvMixer2d(16, 14, 18)
    y = m(_x)
    print(y.shape)



import torch
from torch import nn

from neural_commons.helpers.torch_helper import dims_from
from neural_commons.modules import RndProjection


class InvGaussianLoss(nn.Module):
    def forward(self, pred: torch.Tensor, target: torch.Tensor):
        diff_sq = (pred - target).pow(2)
        mse = torch.mean(diff_sq)
        return 1.0 - torch.exp(-mse * 2)


class IrregularLoss(nn.Module):
    def __init__(self, num_features: int):
        super().__init__()
        self.rp = RndProjection(num_features, 2, seed=3001)

    def forward(self, pred: torch.Tensor, target: torch.Tensor):
        logits = self.rp(target) * 3
        w1 = torch.sigmoid(logits[:, [0]])
        w2 = torch.sigmoid(logits[:, [1]])
        diff_sq = (pred - target).pow(2)
        abs_diff = torch.abs(pred - target)
        dim = dims_from(pred)
        mse = torch.mean(diff_sq, dim=dim)
        loss1 = 1.0 - torch.exp(-mse * w2 * 4)
        loss2 = torch.mean(abs_diff, dim=dim)
        loss = loss1 * w1 + loss2 * (1 - w1)
        return torch.mean(loss)




import torch
from torch import nn


class GaussianNoise(nn.Module):
    def __init__(self, std: float = 0.01):
        super().__init__()
        self.std = std

    def forward(self, x: torch.Tensor):
        if self.training:
            return x + torch.randn_like(x) * self.std
        return x



from typing import Optional, Sequence
import torch
from torch import nn, autograd

from neural_commons.helpers.torch_helper import rms


class ParamModule(nn.Module):
    gradients: Optional[tuple[torch.Tensor, ...]]
    lr: Optional[torch.Tensor]

    def __init__(self, tensors: Sequence[torch.Tensor],
                 approx_lr: float = None,
                 alpha: float = 0.3):
        super().__init__()
        num_tensors = len(tensors)
        if num_tensors == 0:
            raise ValueError("At least one tensor must be provided.")
        for i, t in enumerate(tensors):
            self.register_buffer(f"tensor{i}", t.requires_grad_())
        first_tensor = tensors[0]
        self.num_tensors = num_tensors
        self.gradients = None
        self.lr = None
        self.alpha = alpha
        self.approx_lr_ma = approx_lr or 0.5 * first_tensor.size(0)

    def update_approx_lr(self, lr: float):
        self.approx_lr_ma = lr * self.alpha + self.approx_lr_ma * (1 - self.alpha)

    def get_approx_lr(self) -> float:
        return self.approx_lr_ma

    def set_gradient(self, gradients: tuple[torch.Tensor, ...], lr: torch.Tensor):
        self.gradients = gradients
        self.lr = lr

    def clear_gradient(self):
        self.gradients = None
        self.lr = None

    @property
    def tensors(self) -> tuple[torch.Tensor, ...]:
        return tuple(getattr(self, f"tensor{i}") for i in range(self.num_tensors))

    def increment_tensors(self, offsets: Sequence[torch.Tensor]):
        for i in range(self.num_tensors):
            attr = f"tensor{i}"
            tensor = getattr(self, attr)
            setattr(self, attr, (tensor + offsets[i]).detach().requires_grad_())

    def grad(self, loss: torch.Tensor) -> tuple[torch.Tensor, ...]:
        return autograd.grad(loss, self.tensors)

    def forward(self) -> tuple[torch.Tensor, ...]:
        if self.gradients is not None:
            lr = self.lr
            gen = (t - g * lr for t, g in zip(self.tensors, self.gradients))
            return tuple(gen)
        else:
            return self.tensors



import math
from typing import Iterable, Protocol, Union, Optional
import torch
from torch import autograd, nn

from neural_commons.cf_nn import CFModule
from neural_commons.helpers.torch_helper import concatenate_tensors, rms


class _TLossFn1(Protocol):
    def __call__(self) -> torch.Tensor:
        ...


class _TLossFn2(Protocol):
    def __call__(self, **kwargs) -> torch.Tensor:
        ...


_TLossFn = Union[_TLossFn1, _TLossFn2]


class InvalidCFStateException(BaseException):
    def __init__(self, msg: str):
        super().__init__(msg)


class CFOptimizer:
    def __init__(self, cf_modules: Iterable[CFModule], stochastic_selection: bool = False):
        self.cf_modules = list(cf_modules)
        if len(self.cf_modules) == 0:
            raise ValueError("Empty collection of ParamModules provided.")
        self.selected_index = -1
        self.selected_module: Optional[CFModule] = None
        self.stochastic_selection = stochastic_selection
        self.data_queue: list[tuple[tuple[torch.Tensor, ...], torch.Tensor, torch.Tensor]] = list()
        self._select_next()

    def _full_loss(self, base_loss: torch.Tensor, output: torch.Tensor, std_lambda: float):
        if not self.selected_module.norm_output:
            return base_loss
        extra_loss = (torch.std(output) - 1.0).pow(2) * std_lambda
        return base_loss + extra_loss

    def _grad_pass(self, fwd_fn: _TLossFn, std_lambda: float,
                   **kwargs) -> tuple[torch.Tensor, list[tuple[tuple[torch.Tensor, ...], torch.Tensor, torch.Tensor]]]:
        # Returns list of (input, output, grad)
        inout_list = list()

        def _hook(m: nn.Module, _inputs: tuple[torch.Tensor, ...], _output: torch.Tensor):
            new_output = _output.requires_grad_()
            _in_tuple = tuple(t.detach() for t in _inputs)
            inout_list.append((_in_tuple, new_output,))
            return new_output

        sm = self.selected_module
        h = sm.register_forward_hook(_hook)
        try:
            loss = fwd_fn(**kwargs)
        finally:
            h.remove()

        result = list()
        for inputs, output in inout_list:
            full_loss = self._full_loss(loss, output, std_lambda)
            grad1 = autograd.grad(full_loss, output)[0]
            result.append((inputs, output.detach(), grad1.detach()))
        return loss, result,

    def _lr_loss_grad(self, fwd_fn: _TLossFn, gradients: list[torch.Tensor],
                      lr: float, std_lambda: float, **kwargs) -> tuple[float, float]:
        # Returns loss and LR gradient
        gradient_count = len(gradients)
        lr_t_list = list()

        def _hook(m: nn.Module, _inputs: tuple[torch.Tensor, ...], _output: torch.Tensor):
            index = len(lr_t_list)
            if index >= gradient_count:
                raise InvalidCFStateException("List of gradients obtained in prior forward "
                                              "pass is shorter than the number of module "
                                              "visits in current pass.")
            lr_t = torch.tensor(lr, dtype=torch.float, device=_output.device).requires_grad_()
            new_output = _output - lr_t * gradients[index]
            lr_t_list.append((new_output, lr_t,))
            return new_output

        h = self.selected_module.register_forward_hook(_hook)
        try:
            loss = fwd_fn(**kwargs)
        finally:
            h.remove()

        lr_count = len(lr_t_list)
        if lr_count != gradient_count:
            raise InvalidCFStateException("The number of module visits in current forward pass "
                                          "does not match those of prior pass.")
        sum_grad = 0
        full_loss_sum = 0
        for output, lr_tensor in lr_t_list:
            full_loss = self._full_loss(loss, output, std_lambda)
            grad1 = autograd.grad(full_loss, lr_tensor)[0]
            sum_grad += grad1.item()
            full_loss_sum += full_loss.item()
        return full_loss_sum / lr_count, sum_grad / lr_count,

    def _simple_loss(self, fwd_fn: _TLossFn, gradients: list[torch.Tensor],
                     lr: float, std_lambda: float, **kwargs) -> float:
        gradient_count = len(gradients)
        out_list = list()

        def _hook(m: nn.Module, _inputs: tuple[torch.Tensor, ...], _output: torch.Tensor):
            index = len(out_list)
            if index >= gradient_count:
                raise InvalidCFStateException("List of gradients obtained in prior forward "
                                              "pass is shorter than the number of module "
                                              "visits in current pass.")
            new_output = _output - lr * gradients[index]
            out_list.append(new_output)
            return new_output

        with torch.no_grad():
            h = self.selected_module.register_forward_hook(_hook)
            try:
                loss = fwd_fn(**kwargs)
            finally:
                h.remove()
            num_outputs = len(out_list)
            if num_outputs != gradient_count:
                raise InvalidCFStateException("The number of module visits in current forward pass "
                                              "does not match those of prior pass.")
            full_loss_sum = 0
            for output in out_list:
                full_loss = self._full_loss(loss, output, std_lambda)
                full_loss_sum += full_loss.item()
            return full_loss_sum / num_outputs

    @staticmethod
    def _optimal_lr(lr1: float, lr2: float, lr_loss1: float, lr_loss2: float,
                    lr_grad1: float, lr_grad2: float, lr_expansion=2.0,
                    grad_reduction_threshold=0.03) -> tuple[float, Optional[float]]:
        if math.isinf(lr_grad2) or math.isnan(lr_grad2):
            return lr1, 1.0 / lr_expansion,
        elif math.copysign(1, lr_grad1) != math.copysign(1, lr_grad2):
            # Secant root-finding method
            return (lr1 * lr_grad2 - lr2 * lr_grad1) / (lr_grad2 - lr_grad1), None,
        elif lr_loss2 == lr_loss1:
            return lr1, 1.0,
        elif lr_loss2 < lr_loss1:
            if abs(lr_grad2) < abs(lr_grad1) * grad_reduction_threshold:
                return lr2, 1.0,
            else:
                return lr2, lr_expansion,
        else:
            return lr1, 1.0 / lr_expansion,

    def _update_sm_lr(self, lr: float, prior_approx_lr: float, approx_lr_expansion: float,
                      out_size: int):
        if approx_lr_expansion is not None:
            new_approx_lr = prior_approx_lr * approx_lr_expansion
        else:
            new_approx_lr = lr
        self.selected_module.update_lr_factor(new_approx_lr / out_size)

    def _select_next(self):
        cf_modules = self.cf_modules
        num_modules = len(cf_modules)
        if self.stochastic_selection:
            m_lr = [m.lr_factor_ma for m in cf_modules]
            m_lr = torch.tensor(m_lr, dtype=torch.float)
            m_weight = torch.ones_like(m_lr)
            self.selected_index = torch.multinomial(m_weight, 1).item()
        else:
            self.selected_index = (self.selected_index + 1) % num_modules
        self.selected_module = cf_modules[self.selected_index]

    def enqueue_data(self, fwd_fn: _TLossFn, std_lambda: float = 0.0001,
                     lr_expansion=2.0, **kwargs) -> float:
        loss, pass_info = self._grad_pass(fwd_fn, std_lambda, **kwargs)
        loss = loss.item()
        gradients = [g for _, _, g in pass_info]
        outputs = [out for _, out, _ in pass_info]
        if len(outputs) == 0:
            raise InvalidCFStateException("CFModule instances not visited in forward pass.")
        sm = self.selected_module
        out_size = torch.numel(outputs[0])
        approx_lr = sm.lr_factor_ma * out_size
        lr1 = 0
        lr2 = approx_lr * 2.0
        lr_loss1, lr_grad1 = self._lr_loss_grad(fwd_fn, gradients, lr1, std_lambda, **kwargs)
        lr_loss2, lr_grad2 = self._lr_loss_grad(fwd_fn, gradients, lr2, std_lambda, **kwargs)
        opt_lr, approx_lr_expansion = self._optimal_lr(lr1, lr2, lr_loss1, lr_loss2, lr_grad1, lr_grad2)
        if approx_lr_expansion is None:
            opt_loss = self._simple_loss(fwd_fn, gradients, opt_lr, std_lambda, **kwargs)
            if opt_loss > lr_loss1:
                # Failed to decrease loss. Reduce approx LR.
                opt_lr = lr1
                approx_lr_expansion = 1.0 / lr_expansion
        self._update_sm_lr(opt_lr, approx_lr, approx_lr_expansion, out_size)
        if opt_lr != 0:
            for inputs, output, gradient in pass_info:
                residual = (-gradient * opt_lr).detach()
                self.data_queue.append((inputs, output, residual,))
        return loss

    @property
    def queued_batch_size(self) -> int:
        return sum(output.size(0) for _, output, _ in self.data_queue)

    def _step(self, update_rate: float, max_residual_rel_rms: float, **kwargs):
        if len(self.data_queue) == 0:
            raise InvalidCFStateException("No data has been queued. Call enqueue_data() before calling step().")
        inputs_list = [i for i, _, _ in self.data_queue]
        output_list = [out for _, out, _ in self.data_queue]
        residual_list = [res for _, _, res in self.data_queue]
        self.data_queue = list()
        with torch.no_grad():
            inputs = concatenate_tensors(inputs_list, dim=0)
            del inputs_list
            output = torch.cat(output_list, dim=0)
            del output_list
            residual = torch.cat(residual_list, dim=0)
            del residual_list
            residual *= update_rate
            max_res_rms = rms(output) * max_residual_rel_rms
            res_rms = rms(residual)
            if res_rms > max_res_rms:
                residual *= max_res_rms / res_rms
            self.selected_module.cf_learn(inputs, output, residual, **kwargs)

    def step(self, update_rate: float = 1.0, max_residual_rel_rms = 1.0, **kwargs):
        if update_rate <= 0 or update_rate > 1.0:
            raise ValueError("update_rate expected to be in ]0, 1].")
        try:
            self._step(update_rate, max_residual_rel_rms, **kwargs)
        finally:
            self.data_queue = list()
            self._select_next()



import logging

import torch
from torch import nn


class LogShape(nn.Module):
    def __init__(self, label: str, use_print: bool = False):
        super().__init__()
        self.use_print = use_print
        self.label = label

    def forward(self, x: torch.Tensor):
        message = f"{self.label}: {tuple(x.shape)}"
        if self.use_print:
            print(message)
        else:
            logging.warning(message)
        return x



import torch
from torch import nn
import torch.nn.functional as F


class GaELU(nn.Module):
    def __init__(self):
        super(GaELU, self).__init__()

    def forward(self, x: torch.Tensor):
        p = torch.exp(-(x ** 2))
        return 1.23 * (F.elu(x) - p * 0.3)



from typing import Iterable

import torch

from .CFModule import CFModule
from .CFLinear import CFLinear
from .CFConv2d import CFConv2d
from .CFOptimizer import CFOptimizer


def _cf_modules(self: torch.nn.Module) -> Iterable['CFModule']:
    return (m for m in self.modules() if isinstance(m, CFModule))


torch.nn.Module.cf_modules = _cf_modules



import logging
import math
from typing import Iterable, List, Union, Iterator

import torch
from torch import nn

_pulse = math.pi / 2.0


class CosineScheduleOptimizer:
    def __init__(self, parameters: Union[Iterator[nn.Parameter], List[dict]], lr: float, total_steps: int,
                 min_lr_factor: float = 0.02, warmup_fraction: float = 0.03):
        actual_total_steps = total_steps + 1
        num_warmup_steps = actual_total_steps * warmup_fraction
        warned_step: bool = False

        def _get_lr(step: int):
            nonlocal warned_step
            step_x = step + 0.5
            if step_x < num_warmup_steps:
                current_factor = step_x / num_warmup_steps
            else:
                current_factor = (min_lr_factor +
                                  (1 - min_lr_factor) * math.cos(_pulse * (step_x - num_warmup_steps) /
                                                                 max(1.0, actual_total_steps - num_warmup_steps)))
                if current_factor < 0:
                    current_factor = 0
                    if not warned_step:
                        logging.warning(f'LR Scheduler: step={step}, total_steps={total_steps}')
                        warned_step = True
            return current_factor

        self.optimizer = torch.optim.Adam(parameters, lr=lr)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer,
                                                           lr_lambda=_get_lr)

    def zero_grad(self, set_to_none: bool = False):
        self.optimizer.zero_grad(set_to_none=set_to_none)

    def step(self):
        self.optimizer.step()
        self.scheduler.step()

    def get_lr(self, index: int = 0):
        return self.scheduler.get_last_lr()[index]

    def get_all_lrs(self) -> List[float]:
        return self.scheduler.get_last_lr()



import torch
from torch import nn


class View(nn.Module):
    def __init__(self, *dims: int):
        super().__init__()
        self.dims = dims

    def forward(self, x: torch.Tensor):
        return x.view(self.dims)



import math
import torch
import torch.nn.functional as F
from dataclasses import dataclass
from huggingface_hub import PyTorchModelHubMixin
from torch import nn, Tensor
from torch.nn import Parameter, init


@dataclass
class NeuralLayerConfig:
    in_features: int
    out_features: int
    bias: bool = True
    dtype: str = "float32"
    slopes_log_std: float = 0.1


class NeuralLayer(nn.Module, PyTorchModelHubMixin):
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: Tensor
    hp_params: Tensor

    def __init__(self, in_features: int, out_features: int, bias: bool = True,
                 slopes_log_std: float = 0.1) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        slopes_t = torch.exp(torch.randn(self.out_features) * slopes_log_std)
        self.slopes = nn.Parameter(slopes_t.detach())
        setattr(self.slopes, "__skip_transplant__", True)
        self.weight = Parameter(torch.empty((self.out_features, self.in_features,)))
        if bias:
            self.bias = Parameter(torch.empty(self.out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor) -> Tensor:
        x = F.linear(x, self.weight, self.bias)
        positive = x >= 0
        return positive * x + (~positive) * x * self.slopes[None, :]

    def extra_repr(self) -> str:
        return f'in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}'





import logging
import math
from typing import Iterable, List, Union, Iterator

import torch
from torch import nn

_pulse = math.pi / 2.0


class CosineScheduler:
    def __init__(self, init_value: float, final_value: float, total_steps: int,
                 warmup_fraction: float = 0.03):
        self.final_value = final_value
        self.init_value = init_value
        self._actual_total_steps = total_steps + 1
        self._num_warmup_steps = self._actual_total_steps * warmup_fraction
        self._warned_step: bool = False
        self._index = 0

    def step(self):
        self._index += 1

    @property
    def value(self) -> float:
        step_x = self._index + 0.5
        if step_x < self._num_warmup_steps:
            current_factor = step_x / self._num_warmup_steps
        else:
            current_factor = math.cos(_pulse * (step_x - self._num_warmup_steps) /
                                      max(1.0, self._actual_total_steps -
                                          self._num_warmup_steps))
            if current_factor < 0:
                current_factor = 0
                if not self._warned_step:
                    logging.warning(f'CosineScheduler: step={self._index}, total_steps={self._actual_total_steps}')
                    self._warned_step = True
        return current_factor * self.init_value + (1 - current_factor) * self.final_value



import torch
import torch.nn.functional as F
from torch import nn, Generator


class RndProjection(nn.Module):
    weight: torch.Tensor

    def __init__(self, in_features: int, out_features: int, seed: int = 1001):
        super().__init__()
        weight = self.build_weight_matrix(in_features, out_features, seed=seed)
        self.register_buffer('weight', weight)

    @staticmethod
    def build_weight_matrix(in_features: int, out_features: int, seed: int = 1001):
        # Needs to build a matrix of shape (out_features, in_features,)
        # Given input of shape (b, in_features), multiply by the transpose, and get (b, out_features,)
        return RndProjection._build_weight_impl(in_features, out_features, seed=seed)

    @staticmethod
    def _build_weight_impl(in_features: int, out_features: int, seed: int):
        g = Generator()
        g.manual_seed(seed)
        in_count = 0
        q_list = []
        while in_count < in_features:
            rnd_matrix = torch.randn((out_features, out_features), generator=g)
            q, _ = torch.linalg.qr(rnd_matrix, mode='complete')
            in_count += out_features
            q_list.append(q)
        q_t = torch.cat(q_list, dim=1)
        q_t = torch.clone(q_t[:, :in_features])
        return q_t

    def forward(self, x: torch.Tensor):
        return F.linear(x, self.weight, bias=None)


if __name__ == '__main__':
    torch.set_grad_enabled(True)
    _ol = RndProjection(384, 512)
    _x = torch.randn((5, 384))
    _y = _ol(_x)
    print(_y.size())
    print(torch.std(_y).item())
    print('_x norm: ', torch.norm(_x, dim=1))
    print('_y norm: ', torch.norm(_y, dim=1))





import math

import torch


class AdaptiveSimpleNorm(torch.nn.Module):
    def __init__(self, momentum=0.1, eps=1e-10):
        super(AdaptiveSimpleNorm, self).__init__()
        self.momentum = momentum
        self.eps = eps
        self.mean = 0
        self.std = 1.0

    def forward(self, x):
        if self.training:
            x_mean = torch.mean(x).item()
            x_var = torch.mean((x - self.mean).pow(2)).item()
            x_std = math.sqrt(x_var + self.eps)
            m = self.momentum
            m1m = 1 - m
            self.mean = m1m * self.mean + m * x_mean
            self.std = m1m * self.std + m * x_std
        return (x - self.mean) / self.std



import math
from typing import Union
import torch
import torch.nn.functional as F
from neural_commons.cf_nn import CFModule
from neural_commons.helpers.torch_helper import ridge_regression

_size_2_t = Union[int, tuple[int, int]]
_sqrt_2 = math.sqrt(2)


class CFConv2d(CFModule):
    weight: torch.Tensor
    bias: torch.Tensor

    def __init__(self, in_channels: int, out_channels: int,
                 kernel_size: _size_2_t,
                 stride: _size_2_t = 1,
                 padding: _size_2_t = 0,
                 dilation: _size_2_t = 1,
                 bias: bool = True,
                 norm_output: bool = True,
                 ):
        super().__init__(norm_output=norm_output)
        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size,)
        if isinstance(stride, int):
            stride = (stride, stride,)
        if isinstance(dilation, int):
            dilation = (dilation, dilation,)
        bias_param = None if not bias else torch.randn((out_channels,)) / _sqrt_2
        scale = math.sqrt(in_channels * math.prod(kernel_size))
        if bias:
            scale *= _sqrt_2
        weight = torch.randn((out_channels, in_channels, *kernel_size)) / scale
        self.register_buffer("weight", weight)
        self.register_buffer("bias", bias_param)
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.padding = padding

    def forward(self, x: torch.Tensor):
        return torch.conv2d(x, self.weight, self.bias, self.stride, self.padding, self.dilation,
                            groups=1)

    def cf_learn(self, inputs: tuple[torch.Tensor, ...], output: torch.Tensor, residual: torch.Tensor,
                 l2_lambda: float = 1.0, **kwargs):
        has_bias = self.bias is not None
        x = inputs[0]
        # x: (batch_size, in_channels, height, width)
        # residual: (batch_size, out_channels, out_height, out_width)
        batch_size, _, height, width = x.shape
        unfolded_x = F.unfold(x, self.kernel_size, self.dilation, self.padding, self.stride)
        # unfolded_x: (batch_size, 9 * in_channels, height * width // 2)
        unfolded_x = torch.permute(unfolded_x, (0, 2, 1))
        # unfolded_x: (batch_size, out_height * out_width, 9 * in_channels)
        unfolded_dim_1 = unfolded_x.size(1)
        unfolded_x = unfolded_x.contiguous().view(batch_size * unfolded_dim_1, -1)
        # unfolded_x: (batch_size * out_height * out_width, 9 * in_channels)
        weight = self.weight.view(self.weight.size(0), -1)
        # weight: (out_channels, 9 * in_channels)
        unfolded_residual = torch.permute(residual, (0, 2, 3, 1))
        # unfolded_residual: (batch_size, out_height, out_width, out_channels)
        unfolded_residual = unfolded_residual.contiguous().view(batch_size * unfolded_dim_1, -1)
        # unfolded_residual: (batch_size * out_height * out_width, out_channels)
        w, b = ridge_regression(unfolded_x, unfolded_residual, bias=has_bias, l2_lambda=l2_lambda)
        weight = weight + w.t()
        self.weight = weight.view(*self.weight.shape)
        self.bias = None if b is None else self.bias + b



import torch
from torch import nn


class RMSELoss(nn.Module):
    def __init__(self):
        super().__init__()
        self.mse = nn.MSELoss()

    def forward(self, yhat, y):
        return torch.sqrt(self.mse(yhat, y))



from .RMSELoss import RMSELoss
from .MAELoss import MAELoss
from .GaELU import GaELU
from .GTanh import GTanh
from .ArgMax import ArgMax
from .View import View
from .GaussianNoise import GaussianNoise
from .ParametricReLU import ParametricReLU
from .RndProjection import RndProjection
from .Quadratic import Quadratic
from .Permute import Permute
from .AdaptiveSimpleNorm import AdaptiveSimpleNorm
from .NeuralLayer import NeuralLayer
from .NeuralLayer2d import NeuralLayer2d
from .InputEncoder import InputEncoder
from .SMoEMixIn import SMoEMixIn
from .RndNonLinearFunction import RndNonLinearFunction
from .SMoE import SMoE
from .LogShape import LogShape
from .ConvMixer2d import ConvMixer2d



import torch
from abc import abstractmethod
from torch import nn


class CFModule(nn.Module):
    def __init__(self, initial_lr_factor: float = 0.5, max_lr_factor: float = 1.0, alpha: float = 0.5, norm_output: bool = True):
        super().__init__()
        self.max_lr_factor = max_lr_factor
        self.norm_output = norm_output
        self.lr_factor_ma = initial_lr_factor
        self.alpha = alpha

    def update_lr_factor(self, new_lr_factor: float):
        new_lr_factor_ma = new_lr_factor * self.alpha + self.lr_factor_ma * (1 - self.alpha)
        self.lr_factor_ma = min(self.max_lr_factor, new_lr_factor_ma)

    @abstractmethod
    def cf_learn(self, inputs: tuple[torch.Tensor, ...], output: torch.Tensor, residual: torch.Tensor, **kwargs):
        """
        Applies a (closed-form) learning algorithm that changes the parameters of the module instance.
        Args:
            inputs: The inputs provided to the module.
            output: The original output of the module.
            residual: The residuals that should be added to the original output to obtain the desired output.
            **kwargs: Custom parameters of the learning algorithm.
        """
        pass



import torch.nn as nn
from neural_commons.modules import Quadratic


class RndNonLinearFunction(nn.Module):
    def __init__(self, input_size: int, output_size: int):
        super().__init__()
        hidden_size = max(input_size, output_size) * 15
        self.layers = nn.Sequential(
            Quadratic(input_size, hidden_size),
            Quadratic(hidden_size, output_size),
        )

    def forward(self, x):
        return self.layers(x) * 3



import math
from typing import Iterable, Protocol, Union, Optional
import torch
from torch import autograd

from neural_commons.helpers.torch_helper import rms
from neural_commons.q_nn import ParamModule


class _TLossFn1(Protocol):
    def __call__(self) -> torch.Tensor:
        ...


class _TLossFn2(Protocol):
    def __call__(self, **kwargs) -> torch.Tensor:
        ...


_TLossFn = Union[_TLossFn1, _TLossFn2]


class QGRFOptimizer:
    def __init__(self, p_modules: Iterable[ParamModule], stochastic_selection: bool = False):
        self.p_modules = list(p_modules)
        if len(self.p_modules) == 0:
            raise ValueError("Empty collection of ParamModules provided.")
        self.selected_index = -1
        self.stochastic_selection = stochastic_selection

    @classmethod
    def _lr_loss_grad(cls, fwd_fn: _TLossFn, sm: ParamModule, p_grads: tuple[torch.Tensor, ...],
                      lr: float, l1_lambda: float, l2_lambda: float, dev: torch.device,
                      **kwargs) -> tuple[float, float]:
        lr_t = torch.tensor(lr, device=dev, dtype=torch.float).requires_grad_()
        sm.set_gradient(p_grads, lr_t)
        try:
            lr_loss = cls._fwd_loss(fwd_fn, sm, l1_lambda, l2_lambda, **kwargs)
            lr_grad = autograd.grad(lr_loss, lr_t)[0]
            return lr_loss.item(), lr_grad.item(),
        finally:
            sm.clear_gradient()

    @classmethod
    def _simple_loss(cls, fwd_fn: _TLossFn, sm: ParamModule, p_grads: tuple[torch.Tensor, ...],
                     lr: float, l1_lambda: float, l2_lambda: float,
                     dev: torch.device, **kwargs) -> float:
        with torch.no_grad():
            lr_t = torch.tensor(lr, device=dev, dtype=torch.float)
            sm.set_gradient(p_grads, lr_t)
            try:
                lr_loss = cls._fwd_loss(fwd_fn, sm, l1_lambda, l2_lambda, **kwargs)
                return lr_loss.item()
            finally:
                sm.clear_gradient()

    @staticmethod
    def _optimal_lr(lr1: float, lr2: float, lr_loss1: float, lr_loss2: float,
                    lr_grad1: float, lr_grad2: float, lr_expansion: float) -> tuple[float, Optional[float]]:
        if math.isinf(lr_grad2) or math.isnan(lr_grad2):
            return lr1, 1.0 / lr_expansion,
        elif math.copysign(1, lr_grad1) != math.copysign(1, lr_grad2):
            # Secant root-finding method
            return (lr1 * lr_grad2 - lr2 * lr_grad1) / (lr_grad2 - lr_grad1), None,
        elif lr_loss2 == lr_loss1:
            return lr1, 1.0,
        elif lr_loss2 < lr_loss1:
            return lr2, lr_expansion,
        else:
            return lr1, 1.0 / lr_expansion,

    def _update_pm(self, pm: ParamModule, p_grads: tuple[torch.Tensor, ...],
                   lr: float, update_rate: float,
                   prior_approx_lr: float, approx_lr_factor: float,
                   max_change: float):
        if approx_lr_factor is not None:
            new_approx_lr = prior_approx_lr * approx_lr_factor
        else:
            new_approx_lr = lr
        pm.update_approx_lr(new_approx_lr)
        if lr == 0:
            return
        with torch.no_grad():
            eff_lr = lr * update_rate
            increments = []
            m_tensors = pm.tensors
            g: torch.Tensor
            for g, t in zip(p_grads, m_tensors):
                change = -g * eff_lr
                max_change_rms = rms(t) * max_change
                change_rms = rms(change)
                if change_rms > max_change_rms:
                    change *= max_change_rms / change_rms
                increments.append(change)
            pm.increment_tensors(increments)

    @staticmethod
    def _get_extra_loss(sm: ParamModule, l1_lambda: float,
                        l2_lambda: float) -> torch.Tensor:
        fixed_tensors = sm.tensors
        live_tensors = sm()
        elem_count = 0
        sq_value_sums = []
        abs_value_sums = []
        for lt, ft in zip(live_tensors, fixed_tensors):
            elem_count += torch.numel(lt)
            sq_value_sums.append(torch.sum(lt.pow(2)).view(1))
            abs_value_sums.append(torch.sum(torch.abs(lt)).view(1))
        mean_sq_value = torch.sum(torch.cat(sq_value_sums)) / elem_count
        mean_abs_value = torch.sum(torch.cat(abs_value_sums)) / elem_count
        return l1_lambda * mean_abs_value + l2_lambda * mean_sq_value

    @classmethod
    def _fwd_loss(cls, fwd_fn: _TLossFn, sm: ParamModule,
                  l1_lambda: float, l2_lambda: float, **kwargs):
        return fwd_fn(**kwargs) + cls._get_extra_loss(sm, l1_lambda, l2_lambda)

    def get_approx_lr(self, sm: ParamModule):
        return sm.approx_lr_ma

    def _step(self, fwd_fn: _TLossFn, sm: ParamModule, update_rate: float,
              l1_lambda: float, l2_lambda: float, max_change: float,
              lr_expansion: float = 2.0, **kwargs):
        loss = self._fwd_loss(fwd_fn, sm, l1_lambda, l2_lambda, **kwargs)
        dev = loss.device
        p_grads = sm.grad(loss)
        p_grads = tuple(g.detach() for g in p_grads)
        loss = loss.item()
        approx_lr = self.get_approx_lr(sm)
        lr1 = 0
        lr2 = approx_lr * 2.0
        lr_loss1, lr_grad1 = self._lr_loss_grad(fwd_fn, sm, p_grads, lr1,
                                                l1_lambda, l2_lambda,
                                                dev, **kwargs)
        lr_loss2, lr_grad2 = self._lr_loss_grad(fwd_fn, sm, p_grads, lr2,
                                                l1_lambda, l2_lambda,
                                                dev, **kwargs)
        opt_lr, approx_lr_factor = self._optimal_lr(lr1, lr2, lr_loss1, lr_loss2, lr_grad1, lr_grad2,
                                                    lr_expansion=lr_expansion)
        if approx_lr_factor is None:
            if self._simple_loss(fwd_fn, sm, p_grads, opt_lr, l1_lambda, l2_lambda,
                                 dev, **kwargs) > lr_loss1:
                # Failed to decrease loss. Reduce approx LR.
                opt_lr = lr1
                approx_lr_factor = 1.0 / lr_expansion
        self._update_pm(sm, p_grads, opt_lr, update_rate, approx_lr, approx_lr_factor,
                        max_change=max_change)
        return loss

    def step(self, fwd_fn: _TLossFn, update_rate: float = 1.0, l1_lambda: float = 0.001,
             l2_lambda: float = 0.001, max_change: float = 0.3, **kwargs) -> float:
        if update_rate <= 0 or update_rate > 1.0:
            raise ValueError("update_rate expected to be in ]0, 1].")
        p_modules = self.p_modules
        num_modules = len(p_modules)
        if self.stochastic_selection:
            m_lr = [m.approx_lr_ma for m in p_modules]
            m_lr = torch.tensor(m_lr, dtype=torch.float)
            m_weight = torch.log1p(m_lr)
            self.selected_index = torch.multinomial(m_weight, 1).item()
        else:
            self.selected_index = (self.selected_index + 1) % num_modules
        selected_module = p_modules[self.selected_index]
        return self._step(fwd_fn, selected_module,
                          update_rate=update_rate,
                          l1_lambda=l1_lambda, l2_lambda=l2_lambda,
                          max_change=max_change,
                          **kwargs)



from typing import Optional
import math
import torch


class SMoEMixIn:
    def __init__(self):
        super().__init__()

    @staticmethod
    def get_orthogonal_vectors(n: int, in_features: int):
        if n > in_features:
            q: Optional[torch.Tensor] = None
            while q is None or q.size(1) < n:
                q0, _ = torch.linalg.qr(torch.randn((in_features, in_features)))
                if q is None:
                    q = q0
                else:
                    q = torch.cat([q, q0], dim=1)
            q = torch.clone(q[:, :n])
        else:
            q, _ = torch.linalg.qr(torch.randn((in_features, n)))
        q = torch.transpose(q, 0, 1)
        return q * math.sqrt(in_features)



from dataclasses import dataclass
import torch
from huggingface_hub import PyTorchModelHubMixin
from torch import nn
from neural_commons.modules import NeuralLayer, AdaptiveSimpleNorm


@dataclass
class InputEncoderConfig:
    in_features: int
    out_features: int
    hidden_factor: int = 3
    tag: str = ""
    latent_std: float = None


class InputEncoder(nn.Module, PyTorchModelHubMixin):
    def __init__(self, config: InputEncoderConfig):
        super(InputEncoder, self).__init__()
        super(PyTorchModelHubMixin, self).__init__()
        if isinstance(config, dict):
            config = InputEncoderConfig(**config)
        in_features = config.in_features
        out_features = config.out_features
        hidden_factor = config.hidden_factor
        hidden1 = out_features * hidden_factor
        self.config = config
        self.latent_std = config.latent_std
        self.encoder = nn.Sequential(
            NeuralLayer(in_features, hidden1),
            NeuralLayer(hidden1, hidden1),
            nn.Linear(hidden1, out_features),
        )

    def forward(self, x: torch.Tensor):
        y = self.encoder(x)
        if self.latent_std is not None:
            y = y / self.latent_std
        return y



import torch
from typing import Union
from torch import nn, Tensor


class NeuralLayer2d(nn.Module):
    def __init__(self, in_channels: int, out_channels: int,
                 kernel_size: Union[int, tuple[int, int]],
                 slopes_log_std=0.1,
                 **kwargs):
        super().__init__()
        self.conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                              kernel_size=kernel_size, **kwargs)
        slopes_t = torch.exp(torch.randn(out_channels) * slopes_log_std)
        self.slopes = nn.Parameter(slopes_t.detach())
        setattr(self.slopes, "__skip_transplant__", True)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        positive = x >= 0
        return positive * x + (~positive) * x * self.slopes[None, :, None, None]



import math
import torch
import torch.nn.functional as F
from neural_commons.helpers.torch_helper import ridge_regression
from neural_commons.cf_nn import CFModule


_sqrt_2 = math.sqrt(2)


class CFLinear(CFModule):
    __constants__ = ["in_features", "out_features"]
    weight: torch.Tensor
    bias: torch.Tensor

    def __init__(self, in_features: int, out_features: int, bias: bool = True,
                 norm_output: bool = True):
        super().__init__(norm_output=norm_output)
        scale = math.sqrt(in_features)
        if bias:
            scale *= _sqrt_2
        weight = torch.randn((out_features, in_features)) / scale
        bias_param = torch.randn((out_features,)) / _sqrt_2 if bias else None
        self.register_buffer("weight", weight)
        self.register_buffer("bias", bias_param)
        self.in_features = in_features
        self.out_features = out_features

    def forward(self, x: torch.Tensor):
        return F.linear(x, self.weight, self.bias)

    def cf_learn(self, inputs: tuple[torch.Tensor, ...], output: torch.Tensor, residual: torch.Tensor,
                 l2_lambda: float = 1.0, **kwargs):
        has_bias = self.bias is not None
        x = inputs[0]
        w, b = ridge_regression(x, residual, bias=has_bias, l2_lambda=l2_lambda)
        self.weight = self.weight + w.t()
        self.bias = None if b is None else self.bias + b

    def extra_repr(self) -> str:
        return f'in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}'



import torch
from torch import nn


class GTanh(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        p = torch.exp(-(x ** 2))
        t = torch.tanh(x)
        return 2.4 * p + 0.2 * t - 1.6



import math
from typing import Iterable, Optional, Union, Sequence
import numpy as np
import torch
from torch import nn, autograd, vmap
import torch.nn.functional as F

_sqrt_2 = math.sqrt(2)


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def count_parameters_from_list(param_list: Iterable[nn.Parameter]):
    return sum(p.numel() for p in param_list)


def sin_pos_embeddings(seq_len: int, emb_dim: int) -> torch.Tensor:
    factors = torch.arange(1, emb_dim + 1)
    seq_range = torch.arange(0, seq_len)
    angles = math.pi * 2 * seq_range / seq_len
    return torch.sin(angles[:, None] * factors[None, :]) * _sqrt_2


def concatenate_tensors(list_of_tensors, dim: int = 0):
    if isinstance(list_of_tensors[0], tuple):
        concatenated_tuples = []
        for i in range(len(list_of_tensors[0])):
            concatenated_tensors = torch.cat([t[i] for t in list_of_tensors], dim=dim)
            concatenated_tuples.append(concatenated_tensors)
        return tuple(concatenated_tuples)
    else:
        return torch.cat(list_of_tensors, dim=dim)


def batched_apply(model: nn.Module, x: torch.Tensor, batch_size: int = 64, detached: bool = False):
    output_list = []
    num_items = x.size(0)
    for b0 in range(0, num_items, batch_size):
        input_batch = x[b0:b0 + batch_size]
        output_batch = model(input_batch)
        if detached:
            if isinstance(output_batch, tuple):
                output_batch = tuple([x.detach() for x in output_batch])
            else:
                output_batch = output_batch.detach()
        output_list.append(output_batch)
    output = concatenate_tensors(output_list)
    return output


def sample_tensor(tensor: torch.Tensor, count: int):
    perm = torch.randperm(tensor.size(0), device=tensor.device)
    return tensor[perm[:count]]


def sample_tensors(tensor1: torch.Tensor, tensor2: torch.Tensor, count: int):
    if tensor2.size(0) != tensor1.size(0):
        raise ValueError('Different batch sizes!')
    perm = torch.randperm(tensor1.size(0), device=tensor1.device)
    selection = perm[:count]
    return tensor1[selection], tensor2[selection],


def fixed_hash(text: str):
    h = 0
    for ch in text:
        h = (h * 281 ^ ord(ch) * 997) & 0xFFFFFFFF
    return h


def tensor_hash(t: torch.Tensor):
    return fixed_hash(str(t.cpu().tolist()))


def get_torch_dtype(dtype_str):
    if hasattr(torch, dtype_str):
        return getattr(torch, dtype_str)
    else:
        raise ValueError(f"Unknown dtype string: {dtype_str}")


def ridge_regression(x: torch.Tensor, y: torch.Tensor, bias: bool = True,
                     l2_lambda: float = 1e-5) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Performs ridge regression.
    Args:
        x: Input variables.
        y: Output variables.
        bias: Whether a bias parameter should be returned.
        l2_lambda: An L2 regularization parameter.
    Returns: A weight matrix of shape (in_features, out_features), and a bias parameter
    array of size (out_features) if the bias parameter is True.
    """
    with torch.no_grad():
        dev = x.device
        x_actual = x if not bias else torch.cat([x, torch.ones((x.size(0), 1),
                                                               device=dev)], dim=1)
        x_t = x_actual.t()
        sq_matrix = x_t @ x_actual
        weights = torch.inverse(
            sq_matrix + l2_lambda * torch.eye(x_t.size(0), device=dev)) @ x_t @ y
        # weights: (in_features, out_features)
        bias_param = torch.clone(weights[-1, :]) if bias else None
        if bias:
            weights = weights[:-1, :]
        return weights, bias_param,


def detach_tuple(tup: tuple[torch.Tensor, ...]) -> tuple[torch.Tensor, ...]:
    return tuple(t.detach() for t in tup)


def trim_outliers(tensor: torch.Tensor, zero_based: bool = False, z_score_threshold=3.0,
                  two_passes: bool = False, eps=1e-10):
    mean = 0 if zero_based else torch.mean(tensor).item()
    sd = torch.sqrt(torch.mean((tensor - mean).pow(2)))
    z_score = torch.abs(tensor - mean) / torch.clamp(sd, min=eps)
    is_outlier = z_score >= z_score_threshold
    if two_passes:
        sub_tensor = tensor[~is_outlier]
        if torch.numel(sub_tensor) == 0:
            raise ValueError("Unexpected: All elements are outliers.")
        mean = 0 if zero_based else torch.mean(sub_tensor).item()
        sd = torch.sqrt(torch.mean((sub_tensor - mean).pow(2)))
        z_score = torch.abs(tensor - mean) / torch.clamp(sd, min=eps)
        is_outlier = z_score >= z_score_threshold
    tensor_copy = torch.clone(tensor)
    tensor_copy[is_outlier] = mean + torch.sign(tensor_copy[is_outlier] - mean) * sd * z_score_threshold
    return tensor_copy


def ps_grad(outputs: torch.Tensor, inputs: Union[torch.Tensor, Sequence[torch.Tensor]],
            create_graph: bool = False,
            retain_graph: bool = False) -> tuple[torch.Tensor, ...]:
    """
    Per-sample gradients given per-sample outputs/loss.
    """
    return autograd.grad(outputs.sum(), inputs, create_graph=create_graph,
                         retain_graph=retain_graph)


def view_dims(t: torch.Tensor, n_dims: int):
    s = t.size()
    num_existing = len(s)
    if num_existing > n_dims:
        raise ValueError(f"Provided tensor already has more than {n_dims} axes.")
    return t.view(s + (1,) * (n_dims - num_existing))


def dims_from(t: torch.Tensor, start_dim: int = 1):
    n = len(t.size())
    return tuple(range(start_dim, n))


def num_dims(t: torch.Tensor):
    return len(t.size())


def rms(t: torch.Tensor):
    return torch.sqrt(torch.mean(t.pow(2)))


from typing import Union

import torch
from torch import nn


class Plus(nn.Module):
    def __init__(self, offset: Union[torch.Tensor, float]):
        super().__init__()
        self.offset = offset

    def forward(self, x: torch.Tensor):
        return x + self.offset



from typing import Union

import torch
from torch import nn


class ArgMax(nn.Module):
    def __init__(self, dim: Union[int, None] = -1):
        super().__init__()
        self.dim = dim

    def forward(self, x: torch.Tensor):
        return torch.argmax(x, dim=self.dim)
